unit UParser;

{$MODE Delphi}

interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, USourceCode, FileUtil;

type
  TOnErrorMessage = procedure(Text: string; Position: TPoint) of object;

  TTokenType = (ttNone, ttIdentifier, ttConstantNumber, ttConstantString,
    ttOperator, ttEndOfFile, ttLineComment, ttBlockComment1, ttBlockComment2,
    ttUnknown, ttWhiteSpace);

  { TBaseParser }

  TBaseParser = class
  private
    ProgramCode: TProgram;
    FOnErrorMessage: TOnErrorMessage;
    FNextToken: string;
    FNextTokenType: TTokenType;
    PreviousChar: char;
    CurrentChar: char;
    procedure ErrorMessage(const Text: string; const Arguments: array of const);
  public
    CodeStreamPosition: integer;
    CodePosition: TPoint;
    SourceCodeText: TStringList;
    function IsAlphanumeric(Character: char): boolean;
    procedure GetNextToken;
    function ReadCode: string;
    procedure Expect(Code: string);
    function IsWhiteSpace(Character: char): boolean;
    function IsAlphabetic(Character: char): boolean;
    function IsIdentificator(Text: string): boolean;
    function IsKeyword(Text: string): boolean;
    function IsOperator(Text: string): boolean;
    procedure Log(Text: string);
    property OnErrorMessage: TOnErrorMessage read FOnErrorMessage write FOnErrorMessage;
    procedure Init;
  end;

  { TPascalParser }

  TPascalParser = class(TBaseParser)
    procedure ParseWhileDo(SourceCode: TWhileDo);
    procedure ParseExpression(SourceCode: TExpression);
    procedure ParseUsedModuleList(SourceCode: TUsedModuleList);
    procedure ParseModule(SourceCode: TModule);
    procedure ParseUnit(SourceCode: TModule);
    procedure ParseProgram(SourceCode: TModule);
    procedure ParseAll(SourceCode: TProgram);
    procedure ParseCommonBlock(SourceCode: TCommonBlock; EndSymbol: char = ';');
    function ParseCommand(SourceCode: TCommonBlock): TCommand;
    procedure ParseBeginEnd(SourceCode: TBeginEnd);
    procedure ParseFunctionList(SourceCode: TFunctionList);
    procedure ParseIfThenElse(SourceCode: TIfThenElse);
    procedure ParseForToDo(SourceCode: TForToDo);
    procedure ParseVariableList(SourceCode: TVariableList);
    procedure ParseVariable(SourceCode: TVariable);
    procedure ParseConstantList(SourceCode: TConstantList);
    procedure ParseTypeList(SourceCode: TTypeList);
    function ParseType(TypeList: TTypeList; ExpectName: Boolean = True; AssignSymbol: string = '='): TType;
  private
  end;


implementation

resourcestring
  SUnknownIdentifier = 'Unknown identificator "%s".';
  SIllegalExpression = 'Illegal expression "%s".';
  SExpectedButFound = 'Expected "%s" but "%s" found.';
  SRedefineIdentifier = 'Identificator "%s" redefinition.';
  STypeNotDefined = 'Type "%s" not defined.';
  SEndOfDataReached = 'Parser reached to end of input data.';
  SUndefinedVariable = 'Undefined variable "%s".';
  SUndefinedType = 'Undefined type "%s".';
  SUndefinedConstant = 'Undefined constant "%s".';

{ TBaseParser }

procedure TBaseParser.ErrorMessage(const Text: string; const Arguments: array of const);
begin
  if Assigned(FOnErrorMessage) then
    FOnErrorMessage(Format(Text, Arguments), CodePosition);
end;

procedure TBaseParser.Expect(Code: string);
begin
  Log('Expected: ' + Code + '  Readed: ' + FNextToken);
  if FNextToken <> Code then
  begin
    ErrorMessage(SExpectedButFound, [Code, FNextToken]);

    // Recovery: try to find nearest same code
    while (FNextToken <> Code) and (FNextTokenType <> ttEndOfFile) do
      GetNextToken;
  end;
  GetNextToken;
end;

function TBaseParser.IsAlphabetic(Character: char): boolean;
begin
  Result := (Character in ['a'..'z']) or (Character in ['A'..'Z']);
end;

function TBaseParser.IsAlphanumeric(Character: char): boolean;
begin
  Result := IsAlphabetic(Character) or (Character in ['0'..'9']);
end;

function TBaseParser.IsKeyword(Text: string): boolean;
var
  I: integer;
begin
  Result := False;
  for I := 0 to High(Keywords) do
    if Keywords[I] = Text then
      Result := True;
end;

function TBaseParser.IsOperator(Text: string): boolean;
var
  I: integer;
begin
  Result := False;
  for I := 0 to High(Operators) do
    if Operators[I] = Text then
      Result := True;
end;

function TBaseParser.IsIdentificator(Text: string): boolean;
var
  I: integer;
begin
  Result := True;
  if Length(Text) = 0 then
    Result := False;
  if IsKeyWord(Text) then
    Result := False;
  if Length(Text) > 0 then
    if not (Text[1] in ['a'..'z', 'A'..'Z', '%', '_']) then
      Result := False;
  for I := 2 to Length(Text) do
    if not (Text[i] in ['a'..'z', 'A'..'Z', '0'..'9', '_']) then
      Result := False;
end;

function TBaseParser.IsWhiteSpace(Character: char): boolean;
begin
  Result := (Character = ' ') or (Character = #13) or (Character = #10);
end;

procedure TBaseParser.Log(Text: string);
const
  LogFileName = 'ParseLog.txt';
var
  LogFile: TFileStream;
begin
  try
    if FileExistsUTF8(LogFileName) { *Converted from FileExists*  } then
      LogFile := TFileStream.Create(LogFileName, fmOpenWrite)
    else
      LogFile := TFileStream.Create(LogFileName, fmCreate);
    if Length(Text) > 0 then
    begin
      LogFile.Write(Text[1], Length(Text));
      LogFile.Write(#13#10, 2);
    end;
  finally
    LogFile.Free;
  end;
end;

procedure TBaseParser.Init;
begin
  CodePosition := Point(1, 1);
  CurrentChar := #0;
  PreviousChar := #0;
  FNextToken := '';
  FNextTokenType := ttNone;
  CodeStreamPosition := 1;
  GetNextToken;
end;

procedure TBaseParser.GetNextToken;
var
  I: integer;
  II: integer;
  J: integer;
const
  SpecChar: set of char = [';', '.', ',', ':', '(', ')', '[', ']',
    '+', '-', '/', '*', '^', '=', '<', '>', '@'];
  DoubleSpecChar: array[0..6] of string = (':=', '..', '<=', '>=', '<>', '+=', '-=');
begin
  FNextToken := '';
  FNextTokenType := ttNone;
  with SourceCodeText do
    while True do
    begin
      if CodeStreamPosition < Length(Text) then
      begin
        CurrentChar := Text[CodeStreamPosition];
      end
      else
      begin
        FNextToken := '';
        FNextTokenType := ttEndOfFile;
        Break;
      end;

      if FNextTokenType = ttNone then
      begin
        if IsWhiteSpace(CurrentChar) then
          FNextTokenType := ttWhiteSpace
        else
        if CurrentChar = '{' then
        begin
          FNextTokenType := ttBlockComment1;
        end
        else
        if CurrentChar = '''' then
        begin
          FNextTokenType := ttConstantString;
        end
        else
        if CurrentChar in SpecChar then
        begin
          FNextTokenType := ttOperator;
          FNextToken := FNextToken + CurrentChar;
        end
        else
        if IsAlphanumeric(CurrentChar) then
        begin
          FNextTokenType := ttIdentifier;
          FNextToken := FNextToken + CurrentChar;
        end
        else
          FNextTokenType := ttUnknown;
      end
      else
      if FNextTokenType = ttLineComment then
      begin
        if (CurrentChar = #13) or (CurrentChar = #10) then
          FNextTokenType := ttNone;
      end
      else
      if FNextTokenType = ttBlockComment1 then
      begin
        if (CurrentChar = '}') then
          FNextTokenType := ttNone;
      end
      else
      if FNextTokenType = ttBlockComment2 then
      begin
        if (PreviousChar = '*') and (CurrentChar = ')') then
          FNextTokenType := ttNone;
      end
      else
      if FNextTokenType = ttConstantString then
      begin
        if (CurrentChar = '''') and (PreviousChar = '''') then
          Break
        else
          FNextToken := FNextToken + CurrentChar;
      end
      else
      if FNextTokenType = ttOperator then
      begin
        if (CurrentChar = '*') and (PreviousChar = '(') then
        begin
          FNextToken := '';
          FNextTokenType := ttBlockComment2;
        end
        else
        if (CurrentChar = '/') and (PreviousChar = '/') then
        begin
          FNextToken := '';
          FNextTokenType := ttLineComment;
        end
        else
        if not (CurrentChar in SpecChar) then
          Break
        else
        begin
          J := 0;
          while (J < Length(DoubleSpecChar)) and
            ((PreviousChar + CurrentChar) <> DoubleSpecChar[J]) do
            Inc(J);
          if J < Length(DoubleSpecChar) then
            FNextToken := FNextToken + CurrentChar
          else
            Break;
        end;
      end
      else
      if FNextTokenType = ttIdentifier then
      begin
        if not IsAlphanumeric(CurrentChar) then
          Break
        else
          FNextToken := FNextToken + CurrentChar;
      end
      else if FNextTokenType = ttWhiteSpace then
        FNextTokenType := ttNone;

      if FNextTokenType <> ttNone then
      begin
        // Update cursor position
        Inc(CodePosition.X);
        if (CurrentChar = #13) then
        begin
          CodePosition.X := 1;
          Inc(CodePosition.Y);
        end;

        Inc(CodeStreamPosition);
        PreviousChar := CurrentChar;
      end;
    end;
end;

function TBaseParser.ReadCode: string;
begin
  Result := FNextToken;
  GetNextToken;
  Log('Read: ' + Result);
end;

{ TParserWhileDo }

procedure TPascalParser.ParseWhileDo(SourceCode: TWhileDo);
begin
  with SourceCode do
  begin
    Expect('while');
    Condition.CommonBlock := CommonBlock;
    ParseExpression(Condition);
    Expect('do');
    Command := ParseCommand(CommonBlock);
  end;
end;

{ TExpression }

procedure TPascalParser.ParseExpression(SourceCode: TExpression);
var
  Identifier: string;
  NewVariable: TVariable;
  NewExpression: TExpression;
  Method: TFunction;
  Constant: TConstant;
  //  Brackets: Integer;
  Expressions: TExpressionList;
  I: integer;
  II: integer;
begin
  Expressions := TExpressionList.Create;
  Expressions.Add(TExpression.Create);
  with SourceCode do
  begin
    while ((FNextToken <> ';') and (FNextToken <> ',') and
        (not IsKeyWord(FNextToken))) and not
      (((FNextToken = ')') or (FNextToken = ']'))) do
    begin
      Identifier := ReadCode;
      if Identifier = '(' then
      begin
        // Subexpression
        with TExpression(Expressions.Last) do
        begin
          SubItems[1] := TExpression.Create;
          ParseExpression(TExpression(SubItems[1]));
        end;
        with TExpression(Expressions.Items[Expressions.Add(TExpression.Create)]) do
        begin
          CommonBlock := SourceCode.CommonBlock;
          SubItems[0] := TExpression(Expressions[Expressions.Count - 2]).SubItems[1];
        end;
        Expect(')');
      end
      else
      if IsOperator(Identifier) then
      begin
        // Operator
        TExpression(Expressions.Last).OperatorName := Identifier;
        TExpression(Expressions.Last).NodeType := ntOperator;
      end
      else
      if IsIdentificator(Identifier) then
      begin
        // Reference to identificator
        NewVariable := CommonBlock.Variables.Search(Identifier);
        if Assigned(NewVariable) then
        begin
          // Referenced variable
          with TExpression(Expressions.Last) do
          begin
            SubItems[1] := TExpression.Create;
            TExpression(SubItems[1]).NodeType := ntVariable;
            TExpression(SubItems[1]).Variable := NewVariable;
          end;
          with TExpression(Expressions.Items[Expressions.Add(TExpression.Create)]) do
          begin
            CommonBlock := SourceCode.CommonBlock;
            SubItems[0] := TExpression(Expressions[Expressions.Count - 2]).SubItems[1];
          end;
        end
        else
        begin
          Method := CommonBlock.Functions.Search(Identifier);
          if Assigned(Method) then
          begin
            // Referenced method
            with TExpression(Expressions.Last) do
            begin
              SubItems[1] := TExpression.Create;
              if FNextToken = '(' then               // Method with parameters
                with TExpression(SubItems[1]) do
                begin
                  Expect('(');
                  NewExpression := TExpression.Create;
                  NewExpression.CommonBlock := CommonBlock;
                  ParseExpression(NewExpression);
                  SubItems.Add(NewExpression);
                  while FNextToken = ',' do
                  begin
                    Expect(',');
                    NewExpression := TExpression.Create;
                    NewExpression.CommonBlock := CommonBlock;
                    ParseExpression(NewExpression);
                    SubItems.Add(NewExpression);
                  end;
                  Expect(')');
                end;
              TExpression(SubItems[1]).NodeType := ntFunction;
              TExpression(SubItems[1]).Method := Method;
            end;
            with TExpression(Expressions.Items[Expressions.Add(TExpression.Create)]) do
            begin
              CommonBlock := SourceCode.CommonBlock;
              SubItems[0] :=
                TExpression(Expressions[Expressions.Count - 2]).SubItems[1];
            end;
          end
          else
          begin
            Constant := CommonBlock.Constants.Search(Identifier);
            if Assigned(Constant) then
            begin
              // Referenced constant
              with TExpression(Expressions.Last) do
              begin
                SubItems[1] := TExpression.Create;
                TExpression(SubItems[1]).NodeType := ntConstant;
                TExpression(SubItems[1]).Value := Constant.Value;
              end;
              with TExpression(Expressions.Items[Expressions.Add(
                  TExpression.Create)]) do
              begin
                CommonBlock := SourceCode.CommonBlock;
                SubItems[0] :=
                  TExpression(Expressions[Expressions.Count - 2]).SubItems[1];
              end;
            end
            else
            begin
              ErrorMessage(SUnknownIdentifier, [Identifier]);
            end;
          end;
        end;
      end
      else
      begin
        // Constant value
        with TExpression(Expressions.Last) do
        begin
          SubItems[1] := TExpression.Create;
          TExpression(SubItems[1]).CommonBlock := SourceCode.CommonBlock;
          TExpression(SubItems[1]).NodeType := ntConstant;

          if Identifier[1] = '''' then
          begin
            TExpression(SubItems[1]).Value := Identifier;
            //SetLength(TExpression(SubItems[1]).Value, Length(Identifier));
            //for I := 1 to Length(Identifier) do
            //  TExpression(SubItems[1]).Value[I - 1] := Byte(Identifier[I]);
          end
          else
          begin
            TExpression(SubItems[1]).Value := Identifier;
          end;
        end;
        //ShowMessage(IntToStr(Expressions.Count));
        with TExpression(Expressions.Items[Expressions.Add(TExpression.Create)]) do
        begin
          CommonBlock := SourceCode.CommonBlock;
          SubItems[0] := TExpression(Expressions[Expressions.Count - 2]).SubItems[1];
        end;
      end;
    end;

    // Build expression tree
    for II := 0 to High(Operators) do
    begin
      I := 1;
      while (I < Expressions.Count - 1) do
      begin
        if not TExpression(Expressions[I]).Associated and
          (TExpression(Expressions[I]).OperatorName = Operators[II]) then
        begin
          TExpression(Expressions[I]).Associated := True;
          TExpression(Expressions[I - 1]).SubItems[1] := Expressions[I];
          TExpression(Expressions[I + 1]).SubItems[0] := Expressions[I];
          //Expressions.Delete(I);
        end
        else
          Inc(I);
      end;
    end;
    if Assigned(TExpression(Expressions.First).SubItems[1]) then
      Assign(TExpression(TExpression(Expressions.First).SubItems[1]));
    TExpression(Expressions.First).SubItems[1] := nil;
    //ShowMessage(IntToStr(Expressions.Count));
    if Expressions.Count > 1 then
      TExpression(Expressions[1]).SubItems[0] := nil;
    Expressions.Destroy;
  end;
end;

function TPascalParser.ParseCommand(SourceCode: TCommonBlock): TCommand;
var
  Identifier: string;
  Variable: TVariable;
  Method: TFunction;
  First: TOperation;
  Second: TOperation;
  StartIndex: integer;
  LoopVariable: TVariable;
  IdentName: string;
  FunctionName: string;
begin
  begin
    if FNextToken = 'begin' then
    begin
      Result := TBeginEnd.Create;
      TBeginEnd(Result).CommonBlock := SourceCode;
      //ShowMessage(IntToStr(Integer(SourceCode))
      // + ' ' + IntToStr(Integer(Result)));
      ParseBeginEnd(TBeginEnd(Result));
    end
    else
    if FNextToken = 'if' then
    begin
      Result := TIfThenElse.Create;
      TIfThenElse(Result).CommonBlock := SourceCode;
      ParseIfThenElse(TIfThenElse(Result));
    end
    else
    if FNextToken = 'while' then
    begin
      Result := TWhileDo.Create;
      TWhileDo(Result).CommonBlock := SourceCode;
      ParseWhileDo(TWhileDo(Result));
    end
    else
    if FNextToken = 'for' then
    begin
      Result := TForToDo.Create;
      TForToDo(Result).CommonBlock := SourceCode;
      ParseForToDo(TForToDo(Result));
    end
    else
    if IsIdentificator(FNextToken) then
    begin
      if Assigned(SourceCode.Variables.Search(FNextToken)) then
      begin
        // Variable assignment
        Result := TAssignment.Create;
        TAssignment(Result).CommonBlock := SourceCode;
        IdentName := ReadCode;
        TAssignment(Result).Target := SourceCode.Variables.Search(IdentName);
        Expect(':=');
        TAssignment(Result).Source := TExpression.Create;
        TAssignment(Result).Source.CommonBlock := SourceCode;
        ParseExpression(TAssignment(Result).Source);
      end
      else
      if Assigned(SourceCode.Functions.Search(FNextToken)) then
      begin
        // Function call
        FunctionName := ReadCode;
        Result := TFunctionCall.Create;
        TFunctionCall(Result).CommonBlock := SourceCode;
        TFunctionCall(Result).FunctionRef := SourceCode.Functions.Search(FunctionName);
        if FNextToken = '(' then
        begin
          Expect('(');
          with TFunctionCall(Result) do
          begin
            ParameterExpression.Add(TExpression.Create);
            TExpression(ParameterExpression.Last).CommonBlock := SourceCode;
            ParseExpression(TExpression(ParameterExpression.Last));
          end;
          Expect(')');
        end;
      end
      else
      begin
        Result := nil;
        ErrorMessage(SUnknownIdentifier, [ReadCode]);
      end;
    end
    else
    if FNextToken = ';' then
    else
    begin
      Result := nil;
      ErrorMessage(SIllegalExpression, [ReadCode]);
    end;
  end;
end;

{ TParserModule }

procedure TPascalParser.ParseModule(SourceCode: TModule);
begin
  if FNextToken = 'program' then
    ParseProgram(SourceCode)
  else if FNextToken = 'unit' then
    ParseUnit(SourceCode)
  else
    ParseProgram(SourceCode);
end;

procedure TPascalParser.ParseProgram(SourceCode: TModule);
var
  Identifier: string;
begin
  with SourceCode do
  begin
    if FNextToken = 'program' then
    begin
      Expect('program');
      Name := ReadCode;
      ModuleType := mdProgram;
      Expect(';');
    end
    else
      Name := '';

    // Uses section
    if FNextToken = 'uses' then
      ParseUsedModuleList(UsedModules);

    ParseCommonBlock(SourceCode, '.');
  end;
end;

procedure TPascalParser.ParseUnit(SourceCode: TModule);
begin
  Expect('unit');
  with TModule(ProgramCode.Modules[0]) do
  begin
    Name := ReadCode;
    ModuleType := mdUnit;
  end;
  Expect(';');
  //ParseInterface;
  //ParseImplementation;
end;

{ TParserProgram }

procedure TPascalParser.ParseAll(SourceCode: TProgram);
var
  I: integer;
begin
  with SourceCode do
  begin
    for I := 0 to Modules.Count - 1 do
      ParseModule(TModule(Modules[I]));
  end;
end;

{ TParserCommonBlock }

procedure TPascalParser.ParseCommonBlock(SourceCode: TCommonBlock;
  EndSymbol: char = ';');
begin
  with SourceCode do
  begin
    while FNextToken <> EndSymbol do
    begin
      if FNextToken = 'var' then
        ParseVariableList(Variables)
      else if FNextToken = 'const' then
        ParseConstantList(Constants)
      else if FNextToken = 'type' then
        ParseTypeList(Types)
      else if FNextToken = 'procedure' then
        ParseFunctionList(Functions)
      else if FNextToken = 'function' then
        ParseFunctionList(Functions)
      else
      begin
        ParseBeginEnd(Code);
        Break;
      end;
    end;
    Expect(EndSymbol);
  end;
end;

{ TParserBeginEnd }

procedure TPascalParser.ParseBeginEnd(SourceCode: TBeginEnd);
var
  NewCommand: TCommand;
begin
  //ShowMessage(IntToStr(Integer(SourceCode)) + ' ' + IntToStr(Integer(SourceCode.CommonBlock)));
  with SourceCode do
  begin
    Expect('begin');
    while (FNextToken <> 'end') and (FNextTokenType <> ttEndOfFile) do
    begin
      NewCommand := ParseCommand(CommonBlock);
      if Assigned(NewCommand) then
        Commands.Add(NewCommand);
      //ShowMessage(NextCode);
      if FNextToken = ';' then
        ReadCode;
    end;
    Expect('end');
  end;
end;

{ TParserParseFunctionList }

procedure TPascalParser.ParseFunctionList(SourceCode: TFunctionList);
var
  Identifiers: TStringList;
  NewValueType: TType;
  TypeName: string;
  VariableName: string;
  Variable: TParameter;
  I: integer;
begin
  Identifiers := TStringList.Create;
  with SourceCode do
  begin
    with TFunction(Items[Add(TFunction.Create)]) do
    begin
      Parent := SourceCode.Parent;
      if FNextToken = 'procedure' then
      begin
        Expect('procedure');
        HaveResult := False;
      end
      else
      begin
        Expect('function');
        HaveResult := True;
      end;
      Name := ReadCode;

      if FNextToken = '(' then
      begin
        Expect('(');
        while FNextToken <> ')' do
        begin
          //    while IsIdentificator(NextCode) do begin
          with TParameterList(Parameters) do
          begin
            VariableName := ReadCode;
            Variable := Search(VariableName);
            if not Assigned(Variable) then
            begin
              Identifiers.Add(VariableName);
              while FNextToken = ',' do
              begin
                Expect(',');
                Identifiers.Add(ReadCode);
              end;
            end
            else
              ErrorMessage(SRedefineIdentifier, [VariableName]);
            Expect(':');
            TypeName := ReadCode;
            NewValueType := Parent.Types.Search(TypeName);
            if not Assigned(NewValueType) then
              ErrorMessage(STypeNotDefined, [TypeName])
            else
              for I := 0 to Identifiers.Count - 1 do
                with TParameter(Items[Add(TParameter.Create)]) do
                begin
                  Name := Identifiers[I];
                  ValueType := NewValueType;
                end;
          end;
        end;
        Expect(')');

        // Parse function result type
        if HaveResult then
        begin
          Expect(':');
          TypeName := ReadCode;
          NewValueType := Parent.Types.Search(TypeName);
          if not Assigned(NewValueType) then
            ErrorMessage(STypeNotDefined, [TypeName])
          else
          begin
            ResultType := NewValueType;
            with TVariable(Parent.Variables.Items[Parent.Variables.Add(
                TVariable.Create)]) do
            begin
              Name := 'Result';
              ValueType := NewValueType;
            end;
          end;
        end;
      end;
    end;
    Expect(';');
    ParseCommonBlock(TFunction(Last));
  end;
  Identifiers.Destroy;
end;

{ TParserIfThenElse }

procedure TPascalParser.ParseIfThenElse(SourceCode: TIfThenElse);
begin
  with Sourcecode do
  begin
    Expect('if');
    Condition.CommonBlock := CommonBlock;
    ParseExpression(Condition);
    Expect('then');
    Command := ParseCommand(CommonBlock);
    if FNextToken = 'else' then
    begin
      Expect('else');
      ElseCommand := ParseCommand(CommonBlock);
    end;
  end;
end;

procedure TPascalParser.ParseForToDo(SourceCode: TForToDo);
var
  VariableName: string;
begin
  with SourceCode do
  begin
    Expect('for');
    VariableName := ReadCode;
    ControlVariable := SourceCode.CommonBlock.Variables.Search(VariableName);
    if not Assigned(ControlVariable) then
      ErrorMessage(SUndefinedVariable, [VariableName]);
    Expect(':=');
    Start.CommonBlock := CommonBlock;
    ParseExpression(Start);
    Expect('to');
    Stop.CommonBlock := CommonBlock;
    ParseExpression(Stop);
    Expect('do');
    Command := ParseCommand(CommonBlock);
  end;
end;

{ TParserVariableList }

procedure TPascalParser.ParseVariableList(SourceCode: TVariableList);
var
  Identifiers: TStringList;
  NewValueType: TType;
  TypeName: string;
  VariableName: string;
  Variable: TVariable;
  I: integer;
begin
  Identifiers := TStringList.Create;
  with SourceCode do
  begin
    Expect('var');
    while IsIdentificator(FNextToken) and (FNextTokenType <> ttEndOfFile) do
    begin
      Identifiers.Clear;
      VariableName := ReadCode;
      Variable := Search(VariableName);
      if not Assigned(Variable) then
      begin
        Identifiers.Add(VariableName);
        while FNextToken = ',' do
        begin
          Expect(',');
          Identifiers.Add(ReadCode);
        end;
      end
      else
        ErrorMessage(SRedefineIdentifier, [VariableName]);
      Expect(':');
      TypeName := ReadCode;
      NewValueType := Parent.Types.Search(TypeName);
      if NewValueType = nil then
        ErrorMessage(STypeNotDefined, [TypeName])
      else
        for I := 0 to Identifiers.Count - 1 do
          with TVariable(Items[Add(TVariable.Create)]) do
          begin
            Name := Identifiers[I];
            ValueType := NewValueType;
          end;
      Expect(';');
    end;
  end;
  Identifiers.Destroy;
end;

{ TParserVariable }

procedure TPascalParser.ParseVariable(SourceCode: TVariable);
begin
  with SourceCode do
  begin
    Name := FNextToken;
    Expect(':=');

  end;
end;

{ TParserConstantList }

procedure TPascalParser.ParseConstantList(SourceCode: TConstantList);
var
  Identifiers: TStringList;
  NewValueType: TType;
  TypeName: string;
  ConstantName: string;
  Constant: TConstant;
  I: integer;
  ConstantValue: string;
begin
  Identifiers := TStringList.Create;
  with SourceCode do
  begin
    Expect('const');
    while IsIdentificator(FNextToken) do
    begin
      ConstantName := ReadCode;
      Constant := Search(ConstantName);
      if not Assigned(Constant) then
      begin
        Identifiers.Add(ConstantName);
        while FNextToken = ',' do
        begin
          Expect(',');
          Identifiers.Add(ReadCode);
        end;
      end
      else
        ErrorMessage(SRedefineIdentifier, [ConstantName]);
      Expect(':');
      TypeName := ReadCode;
      NewValueType := Parent.Types.Search(TypeName);
      Expect('=');
      ConstantValue := ReadCode;
      Expect(';');

      if NewValueType = nil then
        ErrorMessage(STypeNotDefined, [TypeName])
      else
        for I := 0 to Identifiers.Count - 1 do
          with TConstant(Items[Add(TConstant.Create)]) do
          begin
            Name := Identifiers[I];
            ValueType := NewValueType;
            Value := ConstantValue;
          end;
    end;
  end;
  Identifiers.Destroy;
end;

{ TParserTypeList }

procedure TPascalParser.ParseTypeList(SourceCode: TTypeList);
var
  NewType: TType;
begin
  with SourceCode do
  begin
    Expect('type');
    while IsIdentificator(FNextToken) do begin
      NewType := ParseType(SourceCode);
      if Assigned(NewType) then begin
        NewType.Parent := SourceCode;
        Add(NewType);
      end;
      Expect(';');
    end;
  end;
end;

{ TParserType }

function TPascalParser.ParseType(TypeList: TTypeList; ExpectName: Boolean = True; AssignSymbol: string = '='): TType;
var
  Name: string;
  TypeName: string;
begin
  //with SourceCode do
  begin
    if ExpectName then begin
      Name := ReadCode;
      Expect(AssignSymbol);
    end;
    if FNextToken = 'record' then begin
      Expect('record');
      Result := TTypeRecord.Create;
      TTypeRecord(Result).Parent := TypeList;
      TType(Result).Name := Name;
      while (FNextToken <> 'end') and (FNextTokenType <> ttEndOfFile) do
      begin
        TTypeRecord(Result).Items.Add(ParseType(TypeList, True, ':'));
        Expect(';');
      end;
      Expect('end');
    end else
    if FNextToken = 'array' then begin
      Expect('array');
      Result := TTypeArray.Create;
      TTypeArray(Result).Parent := TypeList;
      TType(Result).Name := Name;
      if FNextToken = '[' then begin
        Expect('[');
        TypeName := FNextToken;
        TTypeArray(Result).IndexType := ParseType(TypeList, False);
        if not Assigned(TTypeArray(Result).IndexType) then
          ErrorMessage(SUndefinedType, [TypeName]);
        Expect(']');
      end;
      Expect('of');
      TypeName := FNextToken;
      TTypeArray(Result).ItemType := ParseType(TypeList, False);
      if not Assigned(TTypeArray(Result).ItemType) then
        ErrorMessage(SUndefinedType, [TypeName]);
    end else begin
      TypeName := ReadCode;
      if ExpectName then begin
        Result := TType.Create;
        TType(Result).Parent := TypeList;
        TType(Result).Name := Name;
        TType(Result).UsedType := TypeList.Search(TypeName);
        if not Assigned(TType(Result).UsedType) then
          ErrorMessage(SUndefinedType, [TypeName]);
      end else begin
        TType(Result) := TypeList.Search(TypeName);
        if not Assigned(TType(Result)) then
          ErrorMessage(SUndefinedType, [TypeName]);
      end;
    end;
  end;
end;

{ TParserUsedModuleList }

procedure TPascalParser.ParseUsedModuleList(SourceCode: TUsedModuleList);
var
  NewUsedModule: TUsedModule;
begin
  Expect('uses');
  with TUsedModule(SourceCode.Items[SourceCode.Add(TUsedModule.Create)]) do
  begin
    Name := ReadCode;
  end;
  while FNextToken = ',' do
  begin
    Expect(',');
    with TUsedModule(SourceCode.Items[SourceCode.Add(TUsedModule.Create)]) do
    begin
      Name := ReadCode;
    end;
  end;
  Expect(';');
end;

end.

