unit UProducerC;

{$MODE Delphi}

interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, USourceCode, UProducer, StrUtils;

type

  { TProducerC }

  TProducerC = class(TProducer)
  private
    function TranslateType(Name: string): string;
    function TranslateOperator(Name: string): string;
    procedure Emit(AText: string; NewLine: Boolean = True);
    procedure GenerateUses(UsedModules: TUsedModuleList);
    procedure GenerateModule(Module: TModule);
    procedure GenerateCommonBlock(CommonBlock: TCommonBlock;
      LabelPrefix: string);
    procedure GenerateType(AType: TType);
    procedure GenerateTypes(Types: TTypeList);
    procedure GenerateProgram(ProgramBlock: TProgram);
    procedure GenerateFunctions(Functions: TFunctionList);
    procedure GenerateBeginEnd(BeginEnd: TBeginEnd);
    procedure GenerateCommand(Command: TCommand);
    procedure GenerateWhileDo(WhileDo: TWhileDo);
    procedure GenerateForToDo(ForToDo: TForToDo);
    procedure GenerateIfThenElse(IfThenElse: TIfThenElse);
    procedure GenerateAssignment(Assignment: TAssignment);
    procedure GenerateFunctionCall(FunctionCall: TFunctionCall);
    function GenerateExpression(Expression: TExpression): string;
  public
    TextSource: TStringList;
    IndentationLength: Integer;
    Indetation: Integer;
    procedure Produce; override;
    constructor Create;
    destructor Destroy; override;
  end;

implementation

{ TProducerC }

constructor TProducerC.Create;
begin
  TextSource := TStringList.Create;
  IndentationLength := 2;
end;

destructor TProducerC.Destroy;
begin
  TextSource.Free;
  inherited;
end;

function TProducerC.TranslateType(Name: string): string;
begin
  if Name = 'Byte' then Result := 'uint8'
  else if Name = 'ShortInt' then Result := 'int8'
  else if Name = 'Word' then Result := 'int16'
  else if Name = 'SmallInt' then Result := 'int16'
  else if Name = 'Cardinal' then Result := 'uint32'
  else if Name = 'Integer' then Result := 'int32'
  else if Name = 'Void' then Result := 'void'
  else Result := Name;
end;

function TProducerC.TranslateOperator(Name: string): string;
begin
  if Name = '=' then Result := '=='
  else if Name = 'shl' then Result := '<<'
  else if Name = 'shr' then Result := '>>'
  else if Name = 'not' then Result := '!'
  else if Name = 'mod' then Result := '^'
  else if Name = ':=' then Result := '='
  else if Name = '@' then Result := '*'
  else if Name = 'and' then Result := '&'
  else if Name = 'or' then Result := '|'
  else if Name = 'xor' then Result := '^'
  else Result := Name;
end;

procedure TProducerC.Emit(AText: string; NewLine: Boolean = True);
begin
  with TextSource do begin
    if Count = 0 then Add('');
    if Strings[Count - 1] = '' then
      Strings[Count - 1] := Strings[Count - 1] + DupeString(' ', IndentationLength * Indetation);
    Strings[Count - 1] := Strings[Count - 1] + AText;
    if NewLine then Add('');
  end;
end;

procedure TProducerC.GenerateUses(UsedModules: TUsedModuleList);
var
  I: Integer;
begin
  for I := 0 to UsedModules.Count - 1 do
    Emit('#include "' + TUsedModule(UsedModules[I]).Name + '.h"');
  Emit('');
end;

procedure TProducerC.GenerateModule(Module: TModule);
begin
  Emit('#define int8 char');
  Emit('#define int16 int');
  Emit('#define int32 long');
  Emit('#define uint8 unsigned char');
  Emit('#define uint16 unsigned int');
  Emit('#define uint32 unsigned long');
  Emit('');
  GenerateUses(Module.UsedModules);
  GenerateCommonBlock(Module, '');
end;

procedure TProducerC.Produce;
begin
  inherited;
  TextSource.Clear;
  GenerateProgram(ProgramCode);
end;

procedure TProducerC.GenerateProgram(ProgramBlock: TProgram);
var
  I: Integer;
begin
  Indetation := 0;;
  with ProgramBlock do
  for I := 0 to Modules.Count - 1 do
    GenerateModule(TModule(Modules[I]));
end;

procedure TProducerC.GenerateFunctions(Functions: TFunctionList);
var
  I: Integer;
  J: Integer;
  Line: string;
begin
  for I := 0 to Functions.Count - 1 do
  with TFunction(Functions[I]) do
  if not System then
  begin
    if HaveResult then Line := TranslateType(ResultType.Name) + ' '
      else Line := 'void ';
    Line := Line + Name + '(';
    if Parameters.Count > 0 then
    for J := 0 to Parameters.Count - 1 do begin
      Line := Line + TranslateType(TParameter(Parameters[J]).ValueType.Name) +
        ' ' + TParameter(Parameters[J]).Name;
      if J < Parameters.Count - 1 then Line := Line + ', ';
    end;
    Line := Line + ')';
    Emit(Line);
    GenerateBeginEnd(Code);
    Emit('');
  end;
end;

procedure TProducerC.GenerateBeginEnd(BeginEnd: TBeginEnd);
var
  I: Integer;
begin
  Emit('{');
  Inc(Indetation);


  // Variables
  if BeginEnd.Parent is TCommonBlock then begin
    for I := 0 to BeginEnd.CommonBlock.Variables.Count - 1 do
    with TVariable(BeginEnd.CommonBlock.Variables[I]) do
      Emit(TranslateType(ValueType.Name) + ' ' + Name + ';');
    Emit('');
  end;

  // Commands
  for I := 0 to BeginEnd.Commands.Count - 1 do
    GenerateCommand(TCommand(BeginEnd.Commands[I]));

  Dec(Indetation);
  Emit('}');
end;

procedure TProducerC.GenerateCommand(Command: TCommand);
begin
  if Command is TBeginEnd then GenerateBeginEnd(TBeginEnd(Command))
  else if Command is TWhileDo then GenerateWhileDo(TWhileDo(Command))
  else if Command is TForToDo then GenerateForToDo(TForToDo(Command))
  else if Command is TIfThenElse then GenerateIfThenElse(TIfThenElse(Command))
  else if Command is TAssignment then GenerateAssignment(TAssignment(Command))
  else if Command is TFunctionCall then GenerateFunctionCall(TFunctionCall(Command));
end;

procedure TProducerC.GenerateWhileDo(WhileDo: TWhileDo);
begin
  Emit('while (' + GenerateExpression(WhileDo.Condition) + ')');
  GenerateCommand(WhileDo.Command);
end;

procedure TProducerC.GenerateForToDo(ForToDo: TForToDo);
begin
  with ForToDo do begin
    Emit('for(' + ControlVariable.Name + ' = ' +
      GenerateExpression(Start) + '; ' + ControlVariable.Name + ' < ' +
      GenerateExpression(Stop) + '; ' + ControlVariable.Name + '++)');
    GenerateCommand(Command);
  end;
end;

procedure TProducerC.GenerateIfThenElse(IfThenElse: TIfThenElse);
begin
  Emit('if(' + GenerateExpression(IfThenElse.Condition) + ')');
  GenerateCommand(IfThenElse.Command);
  if Assigned(IfThenElse.ElseCommand) then begin
    Emit('else ');
    GenerateCommand(IfThenElse.ElseCommand);
  end;
end;

procedure TProducerC.GenerateAssignment(Assignment: TAssignment);
begin
  if Assignment.Target.Name = 'Result' then Emit('return(' + GenerateExpression(Assignment.Source) + ');')
    else Emit(Assignment.Target.Name + ' = ' + GenerateExpression(Assignment.Source) + ';');
end;

procedure TProducerC.GenerateFunctionCall(FunctionCall: TFunctionCall);
var
  Line: string;
  I: Integer;
begin
  with FunctionCall do begin
    Line := FunctionRef.Name + '(';
    if ParameterExpression.Count > 0 then begin
      for I := 0 to ParameterExpression.Count - 1 do begin
        Line := Line + GenerateExpression(TExpression(ParameterExpression[I]));
        if I < ParameterExpression.Count - 1 then Line := Line + ', ';
      end;
    end;
    Line := Line + ');';
    Emit(Line);
  end;
end;

function TProducerC.GenerateExpression(Expression: TExpression): string;
begin
  case Expression.NodeType of
    ntConstant: Result := Expression.Value;
    ntVariable: Result := Expression.Variable.Name;
    ntFunction: Result := Expression.Method.Name;
    ntOperator: begin
      Result := GenerateExpression(TExpression(Expression.SubItems.First))
        + ' ' + TranslateOperator(Expression.OperatorName) + ' ' +
        GenerateExpression(TExpression(Expression.SubItems.Last));
    end;
    ntNone: ;
  end;
end;

procedure TProducerC.GenerateCommonBlock(CommonBlock: TCommonBlock; LabelPrefix: string);
var
  I: Integer;
  LabelName: string;
begin
  with CommonBlock do begin
    GenerateTypes(Types);
    GenerateFunctions(Functions);
    Emit('void ' + Name + '()');
    GenerateBeginEnd(Code);
  end;
end;

procedure TProducerC.GenerateType(AType: TType);
var
  I: Integer;
begin
  if AType is TTypeRecord then begin
    Emit('typedef struct');
    Emit('{');
    Inc(Indetation);
    for I := 0 to TTypeRecord(AType).Items.Count - 1 do begin
      GenerateType(TType(TTypeRecord(AType).Items[I]));
      Emit(';');
    end;
    Dec(Indetation);
    Emit('} ' + TranslateType(AType.Name), False);
  end else
  if AType is TTypeArray then begin
    GenerateType(TTypeArray(AType).ItemType);
    Emit('* ', False);

(*    if Assigned(TTypeArray(AType).IndexType) then begin
      Emit(AType.Name + '[', False);
      Emit('[', False);
      GenerateType(TTypeArray(AType).IndexType);
      Emit(']', False);
    end;
    Emit(' of ', False);
    if Assigned(TTypeArray(AType).ItemType) then*)
    Emit(TranslateType(AType.Name), False);
  end else begin
    if Assigned(AType.UsedType) then begin
      GenerateType(AType.UsedType);
      Emit(' ', False);
    end;
    Emit(TranslateType(AType.Name), False);
  end;
end;

procedure TProducerC.GenerateTypes(Types: TTypeList);
var
  I: Integer;
begin
  if Types.Count > 0 then begin
    Inc(Indetation);
    for I := 0 to Types.Count - 1 do
    with TType(Types[I]) do
    if (not System) then begin
      GenerateType(TType(Types[I]));
      Emit(';');
    end;
    Dec(Indetation);
    Emit('');
  end;
end;



end.
