unit UProducerPascal;

{$MODE Delphi}

interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, USourceCode, UProducer, StrUtils;

type

  { TProducerPascal }

  TProducerPascal = class(TProducer)
  private
    procedure Emit(AText: string; NewLine: Boolean = True);
    procedure GenerateUses(UsedModules: TUsedModuleList);
    procedure GenerateModule(Module: TModule);
    procedure GenerateType(AType: TType; AssignSymbol: Char = ':');
    procedure GenerateTypes(Types: TTypeList);
    procedure GenerateCommonBlock(CommonBlock: TCommonBlock;
      LabelPrefix: string);
    procedure GenerateProgram(ProgramBlock: TProgram);
    procedure GenerateFunctions(Functions: TFunctionList);
    procedure GenerateConstants(Constants: TConstantList);
    procedure GenerateBeginEnd(BeginEnd: TBeginEnd);
    procedure GenerateVariableList(Variables: TVariableList);
    procedure GenerateCommand(Command: TCommand);
    procedure GenerateWhileDo(WhileDo: TWhileDo);
    procedure GenerateForToDo(ForToDo: TForToDo);
    procedure GenerateIfThenElse(IfThenElse: TIfThenElse);
    procedure GenerateAssignment(Assignment: TAssignment);
    procedure GenerateFunctionCall(ConstantCall: TFunctionCall);
    function GenerateExpression(Expression: TExpression): string;
  public
    TextSource: TStringList;
    IndentationLength: Integer;
    Indetation: Integer;
    procedure Produce; override;
    constructor Create;
    destructor Destroy; override;
  end;

implementation

{ TProducerPascal }

constructor TProducerPascal.Create;
begin
  TextSource := TStringList.Create;
  IndentationLength := 2;
end;

destructor TProducerPascal.Destroy;
begin
  TextSource.Free;
  inherited;
end;

procedure TProducerPascal.Emit(AText: string; NewLine: Boolean = True);
begin
  with TextSource do begin
    if Count = 0 then Add('');
    if Strings[Count - 1] = '' then
      Strings[Count - 1] := Strings[Count - 1] + DupeString(' ', IndentationLength * Indetation);
    Strings[Count - 1] := Strings[Count - 1] + AText;
    if NewLine then Add('');
  end;
end;

procedure TProducerPascal.GenerateUses(UsedModules: TUsedModuleList);
var
  I: Integer;
  Line: string;
begin
  Line := 'uses ';
  for I := 0 to UsedModules.Count - 1 do begin
    Line := Line + TUsedModule(UsedModules[I]).Name;
    if I < UsedModules.Count - 1 then Line := Line + ', ';
  end;
  Emit(Line + ';');
  Emit('');
end;

procedure TProducerPascal.GenerateModule(Module: TModule);
begin
  Emit('program ' + Module.Name + ';');
  Emit('');
  GenerateUses(Module.UsedModules);
  GenerateCommonBlock(Module, '');
  Emit('.', False);
end;

procedure TProducerPascal.GenerateType(AType: TType; AssignSymbol: Char = ':');
var
  I: Integer;
begin
  if AType is TTypeRecord then begin
    Emit(AType.Name + ' ' + AssignSymbol + ' record');
    Inc(Indetation);
    for I := 0 to TTypeRecord(AType).Items.Count - 1 do begin
      GenerateType(TType(TTypeRecord(AType).Items[I]));
      Emit(';');
    end;
    Dec(Indetation);
    Emit('end', False);
  end else
  if AType is TTypeArray then begin
    Emit(AType.Name + ' ' + AssignSymbol + ' array ', False);
    if Assigned(TTypeArray(AType).IndexType) then begin
      Emit('[', False);
      GenerateType(TTypeArray(AType).IndexType);
      Emit(']', False);
    end;
    Emit(' of ', False);
    if Assigned(TTypeArray(AType).ItemType) then
      GenerateType(TTypeArray(AType).ItemType);
  end else begin
    Emit(AType.Name, False);
    if Assigned(AType.UsedType) then begin
      Emit(' ' + AssignSymbol + ' ', False);
      GenerateType(AType.UsedType);
    end;
  end;
end;

procedure TProducerPascal.GenerateTypes(Types: TTypeList);
var
  I: Integer;
begin
  if Types.Count > 0 then begin
    Emit('type');
    Inc(Indetation);
    for I := 0 to Types.Count - 1 do
    with TType(Types[I]) do
    if (not System) then begin
      GenerateType(TType(Types[I]), '=');
      Emit(';');
    end;
    Dec(Indetation);
    Emit('');
  end;
end;

procedure TProducerPascal.Produce;
begin
  inherited;
  TextSource.Clear;
  GenerateProgram(ProgramCode);
end;

procedure TProducerPascal.GenerateProgram(ProgramBlock: TProgram);
var
  I: Integer;
begin
  Indetation := 0;;
  with ProgramBlock do
  for I := 0 to Modules.Count - 1 do
    GenerateModule(TModule(Modules[I]));
end;

procedure TProducerPascal.GenerateFunctions(Functions: TFunctionList);
var
  I: Integer;
  P: Integer;
  Line: string;
begin
  for I := 0 to Functions.Count - 1 do
  with TFunction(Functions[I]) do
  if not System then
  begin
    if HaveResult then
      Line := 'function ' + Name
      else Line := 'procedure ' + Name;
    if Parameters.Count > 0 then begin
      Line := Line + '(';
      for P := 0 to Parameters.Count - 1 do begin
        with TParameter(Parameters[P]) do
          Line := Line + Name + ': ' + ValueType.Name;
        if P < (Parameters.Count - 1) then Line := Line + '; ';
      end;
      Line := Line + ')';
    end;
    if HaveResult then Line := Line + ': ' + ResultType.Name;
    Emit(Line + ';');
    GenerateBeginEnd(Code);
    Emit(';');
    Emit('');
  end;
end;

procedure TProducerPascal.GenerateConstants(Constants: TConstantList);
var
  I: Integer;
begin
  if Constants.Count > 0 then begin
    Emit('const');
    Inc(Indetation);
    for I := 0 to Constants.Count - 1 do
    with TConstant(Constants[I]) do
    if not System then begin
      Emit(Name + ': ' + ValueType.Name + ' = ' + Value + ';');
    end;
    Dec(Indetation);
    Emit('');
  end;
end;

procedure TProducerPascal.GenerateBeginEnd(BeginEnd: TBeginEnd);
var
  I: Integer;
begin
  Emit('begin');
  Inc(Indetation);
  // Commands
  for I := 0 to BeginEnd.Commands.Count - 1 do begin
    GenerateCommand(TCommand(BeginEnd.Commands[I]));
    Emit(';');
  end;

  Dec(Indetation);
  Emit('end', False);
end;

procedure TProducerPascal.GenerateVariableList(Variables: TVariableList);
var
  I: Integer;
begin
  Emit('var');
  Inc(Indetation);
  for I := 0 to Variables.Count - 1 do
  with TVariable(Variables[I]) do
    Emit(Name + ': ' + ValueType.Name + ';');
  Dec(Indetation);
  Emit('');
end;

procedure TProducerPascal.GenerateCommand(Command: TCommand);
begin
  if Command is TBeginEnd then GenerateBeginEnd(TBeginEnd(Command))
  else if Command is TWhileDo then GenerateWhileDo(TWhileDo(Command))
  else if Command is TForToDo then GenerateForToDo(TForToDo(Command))
  else if Command is TIfThenElse then GenerateIfThenElse(TIfThenElse(Command))
  else if Command is TAssignment then GenerateAssignment(TAssignment(Command))
  else if Command is TFunctionCall then GenerateFunctionCall(TFunctionCall(Command));
end;

procedure TProducerPascal.GenerateWhileDo(WhileDo: TWhileDo);
begin
  Emit('while ' + GenerateExpression(WhileDo.Condition) + ' do ');
  GenerateCommand(WhileDo.Command);
end;

procedure TProducerPascal.GenerateForToDo(ForToDo: TForToDo);
begin
  with ForToDo do begin
    Emit('for ' + ControlVariable.Name + ' := ' +
      GenerateExpression(Start) + ' to ' + GenerateExpression(Stop) + ' do ');
    GenerateCommand(Command);
  end;
end;

procedure TProducerPascal.GenerateIfThenElse(IfThenElse: TIfThenElse);
begin
  Emit('if ' + GenerateExpression(IfThenElse.Condition) + ' then ');
  GenerateCommand(IfThenElse.Command);
  if Assigned(IfThenElse.ElseCommand) then begin
    Emit(' else ');
    GenerateCommand(IfThenElse.ElseCommand);
  end;
end;

procedure TProducerPascal.GenerateAssignment(Assignment: TAssignment);
begin
  Emit(Assignment.Target.Name + ' := ' + GenerateExpression(Assignment.Source), False);
end;

procedure TProducerPascal.GenerateFunctionCall(ConstantCall: TFunctionCall);
var
  Line: string;
  I: Integer;
begin
  with ConstantCall do begin
    Line := FunctionRef.Name;
    if ParameterExpression.Count > 0 then begin
      Line := Line + '(';
      for I := 0 to ParameterExpression.Count - 1 do begin
        Line := Line + GenerateExpression(TExpression(ParameterExpression[I]));
        if I < ParameterExpression.Count - 1 then Line := Line + ', ';
      end;
      Line := Line + ')';
    end;
  end;
  Emit(Line, False);
end;

function TProducerPascal.GenerateExpression(Expression: TExpression): string;
begin
  case Expression.NodeType of
    ntConstant: Result := Expression.Value;
    ntVariable: Result := Expression.Variable.Name;
    ntFunction: Result := Expression.Method.Name;
    ntOperator: begin
      Result := GenerateExpression(TExpression(Expression.SubItems.First))
        + ' ' + Expression.OperatorName + ' ' +
        GenerateExpression(TExpression(Expression.SubItems.Last));
    end;
    ntNone: ;
  end;
end;

procedure TProducerPascal.GenerateCommonBlock(CommonBlock: TCommonBlock; LabelPrefix: string);
var
  I: Integer;
  LabelName: string;
begin
  with CommonBlock do begin
    GenerateTypes(Types);
    GenerateFunctions(Functions);
    GenerateConstants(Constants);
    GenerateVariableList(Variables);
    GenerateBeginEnd(Code);
  end;
end;



end.
