unit UCompiler;

{$MODE Delphi}

interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, USourceCode, UProducer, UParser,
  UProducerC, Contnrs;

type

  TErrorMessage = class
    Text: string;
    Position: TPoint;
  end;

  { TCompiler }

  TCompiler = class
  private
    FOnErrorMessage: TOnErrorMessage;
    procedure ErrorMessage(Text: string; Position: TPoint);
    procedure InitSystem;
  public
    ProgramCode: TProgram;
    SourceCode: TStringList;
    Parser: TPascalParser;
    Producer: TProducer;
    ErrorMessages: TObjectList;
    constructor Create;
    procedure Compile;
    destructor Destroy; override;
    property OnErrorMessage: TOnErrorMessage read FOnErrorMessage write FOnErrorMessage;
  end;

implementation

{ TCompiler }

procedure TCompiler.Compile;
begin
  InitSystem;
  ErrorMessages.Clear;
  Parser.Init;
  Parser.ParseAll(ProgramCode);
  Producer.Produce;
end;

constructor TCompiler.Create;
begin
  SourceCode := TStringList.Create;
  ProgramCode := TProgram.Create;
  Producer := TProducerC.Create;
  Producer.ProgramCode := ProgramCode;
  Parser := TPascalParser.Create;
  Parser.SourceCodeText := SourceCode;
  Parser.OnErrorMessage := ErrorMessage;
  ErrorMessages := TObjectList.Create;
end;

destructor TCompiler.Destroy;
begin
  ProgramCode.Free;
  SourceCode.Free;
  Parser.Free;
  Producer.Free;
  ErrorMessages.Free;
end;

procedure TCompiler.ErrorMessage(Text: string; Position: TPoint);
var
  NewMessage: TErrorMessage;
begin
  NewMessage := TErrorMessage.Create;
  NewMessage.Text := Text;
  NewMessage.Position := Position;
  ErrorMessages.Add(NewMessage);
  if Assigned(FOnErrorMessage) then FOnErrorMessage(Text, Position);
end;

procedure TCompiler.InitSystem;
begin
  with Parser do begin
    Log('==== Parse start ====');
    with ProgramCode do begin
      Modules.Clear;

      with TModule(Modules[Modules.Add(TModule.Create)]) do begin
        Name := 'Main';
        with TType(Types[Types.Add(TType.Create)]) do begin
          Name := 'Void';
          System := True;
          Size := 0;
          UsedType := nil;
        end;
        with TType(Types[Types.Add(TType.Create)]) do begin
          Name := 'Byte';
          System := True;
          Size := 1;
          UsedType := nil;
        end;
        with TType(Types[Types.Add(TType.Create)]) do begin
          Name := 'ShortInt';
          System := True;
          Size := 1;
          UsedType := nil;
        end;
        with TType(Types[Types.Add(TType.Create)]) do begin
          Name := 'Word';
          System := True;
          Size := 2;
          UsedType := nil;
        end;
        with TType(Types[Types.Add(TType.Create)]) do begin
          Name := 'SmallInt';
          System := True;
          Size := 2;
          UsedType := nil;
        end;
        with TType(Types[Types.Add(TType.Create)]) do begin
          Name := 'Cardinal';
          System := True;
          Size := 3;
          UsedType := nil;
        end;
        with TType(Types[Types.Add(TType.Create)]) do begin
          Name := 'Integer';
          System := True;
          Size := 3;
          UsedType := nil;
        end;
        with TType(Types[Types.Add(TType.Create)]) do begin
          Name := 'Char';
          System := True;
          Size := 1;
          UsedType := nil;
        end;
        with TType(Types[Types.Add(TType.Create)]) do begin
          Name := 'String';
          System := True;
          Size := 1;
          UsedType := nil;
        end;
        with TFunction(Functions[Functions.Add(TFunction.Create)]) do begin
          Name := 'Exit';
          System := True;
          ResultType := TType(TModule(Modules[0]).Types[0]);
        end;
        with TFunction(Functions[Functions.Add(TFunction.Create)]) do begin
          Name := 'Break';
          System := True;
          ResultType := TType(TModule(Modules[0]).Types[0]);
        end;
        with TFunction(Functions[Functions.Add(TFunction.Create)]) do begin
          Name := 'Continue';
          System := True;
          ResultType := TType(TModule(Modules[0]).Types[0]);
        end;
        with TFunction(Functions[Functions.Add(TFunction.Create)]) do begin
          Name := 'WriteLn';
          ResultType := TType(TModule(Modules[0]).Types[0]);
          with TParameter(Parameters[Parameters.Add(TParameter.Create)]) do begin
            Name := 'Text';
            ValueType := TType(TModule(Modules[0]).Types[1]);
          end;
        end;
      end;
    end;
  end;
end;

end.
