unit USourceCode;

{$MODE Delphi}

interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Contnrs;

type
  TMemoryType = (mtProgram, mtData, mtEEPROM);

  TModuleType = (mdProgram, mdUnit, mdLibrary, mdPackage);

  TInstruction = (inNone, inJump, inConditionalJump, inExpressionEvaluation,
    inReturn);

  TNodeType = (ntNone, ntVariable, ntFunction, ntConstant, ntOperator);

  TValue = Variant; //array of Byte;

  TCommonBlock = class;
  TTypeList = class;
  TConstantList = class;
  TVariableList = class;
  TFunctionList = class;
  TExpressionList = class;
  TExpression = class;
  TOperationList = class;
  TFunction = class;
  TVariable = class;
  TConstant = class;
  TModule = class;

  TDevice = class
    Family: string;
    Memory: array[TMemoryType] of Integer;
  end;

  TContext = class

  end;

  TCommandList = class;

  TCommand = class
    Parent: TObject;
    CommonBlock: TCommonBlock;
  end;

  TAssignment = class(TCommand)
    Target: TVariable;
    Source: TExpression;
    constructor Create;
    destructor Destroy; override;
  end;

  { TFunctionCall }

  TFunctionCall = class(TCommand)
    FunctionRef: TFunction;
    ParameterExpression: TObjectList; // TObjectList<TExpression>
    constructor Create;
    destructor Destroy; override;
  end;

  TBeginEnd = class(TCommand)
    Commands: TCommandList;
    CommonBlock: TCommonBlock;
    procedure Clear;
    constructor Create;
    destructor Destroy; override;
  end;

  TWhileDo = class(TCommand)
    Condition: TExpression;
    Command: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  TWithDo = class(TCommand)
    Context: TContext;
    Command: TCommand;
  end;

  TRepeatUntil = class(TCommand)
    Block: TCommandList;
    Condition: TExpression;
  end;

  { TForToDo }

  TForToDo = class(TCommand)
    ControlVariable: TVariable;
    Start: TExpression;
    Stop: TExpression;
    Command: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  { TIfThenElse }

  TIfThenElse = class(TCommand)
    Condition: TExpression;
    Command: TCommand;
    ElseCommand: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  TCaseOfEndBranche = class
    Constant: TConstant;
    Command: TCommand;
  end;

  TCaseOfEnd = class(TCommand)
    Expression: TExpression;
    Branches: TObjectList; // TObjectList<TCaseOfEndBranche>
    ElseCommand: TCommand;
    constructor Create;
    destructor Destroy; override;
  end;

  TTryFinally = class(TCommand)
    Block: TCommandList;
    FinallyBlock: TCommandList;
  end;

  TTryExcept = class(TCommand)
    Block: TCommandList;
    ExceptBlock: TCommandList;
  end;

  TCommandList = class(TObjectList)

  end;

  TCommonBlock = class
    Name: string;
    Parent: TCommonBlock;
    Constants: TConstantList;
    Types: TTypeList;
    Variables: TVariableList;
    Functions: TFunctionList;
    Code: TBeginEnd;
    constructor Create; virtual;
    destructor Destroy; override;
//    procedure CheckReferences;
  end;

  TType = class
    System: Boolean;
    Parent: TTypeList;
    Name: string;
    Size: Integer;
    UsedType: TType;
  end;

  { TTypeRecord }

  TTypeRecord = class(TType)
    Items: TObjectList; // TObjectList<TType>
    constructor Create;
    destructor Destroy; override;
  end;

  TTypeArray = class(TType)
    IndexType: TType;
    ItemType: TType;
  end;

  TTypeList = class(TObjectList)
    Parent: TCommonBlock;
    function Search(Name: string): TType;
    destructor Destroy; override;
  end;

  TConstant = class
    System: Boolean;
    Name: string;
    ValueType: TType;
    Value: TValue;
  end;

  TConstantList = class(TObjectList)
    Parent: TCommonBlock;
    function Search(Name: string): TConstant;
    destructor Destroy; override;
  end;

  TVariable = class
    Name: string;
    ValueType: TType;
    DefaultValue: TValue;
  end;

  TVariableList = class(TObjectList)
    Parent: TCommonBlock;
    function Search(Name: string): TVariable;
    destructor Destroy; override;
  end;

  TParameter = class(TVariable)
  end;

  TParameterList = class(TObjectList)
    Parent: TFunction;
    function Search(Name: string): TParameter;
    destructor Destroy; override;
  end;

  { TExpression }

  TExpression = class
    CommonBlock: TCommonBlock;
    NodeType: TNodeType;
    Variable: TVariable;
    Method: TFunction;
    Value: TValue;
    OperatorName: string;
    SubItems: TExpressionList;
    Associated: Boolean;
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TExpression);
  end;

  TExpressionList = class(TObjectList)
    destructor Destroy; override;
  end;

  TOperation = class
    Instruction: TInstruction;
    ExpressionTree: TExpression;
    GotoAddress: Integer;
    Negative: Boolean;
    Referenced: Boolean;
    destructor Destroy; override;
  end;

  TOperationList = class(TObjectList)
    destructor Destroy; override;
  end;

  TFunction = class(TCommonBlock)
  public
    System: Boolean;
    HaveResult: Boolean;
    Parameters: TParameterList;
    ResultType: TType;
    constructor Create; override;
    destructor Destroy; override;
  end;

  TFunctionList = class(TObjectList)
    Parent: TCommonBlock;
    function Search(Name: string): TFunction;
    destructor Destroy; override;
  end;

  TUsedModule = class
    Name: string;
    Module: TModule;
  end;

  TUsedModuleList = class(TObjectList)
  end;

  TModule = class(TCommonBlock)
  public
    ModuleType: TModuleType;
    UsedModules: TUsedModuleList;
    constructor Create; override;
    procedure Clear;
    destructor Destroy; override;
  end;

  TProgram = class
    Device: TDevice;
    Modules: TObjectList; // TObjectList<TModule>
    constructor Create;
    destructor Destroy; override;
  end;

const
  KeyWords: array[0..37] of string = ('program', 'unit', 'uses', 'begin', 'end',
    'type', 'const', 'var', 'array', 'record', 'absolute', 'virtual', 'class',
    'set', 'private', 'public', 'interface', 'implementation', 'finalization',
    'initialization', 'for', 'while', 'if', 'case', 'of', 'pointer',
    'object', 'packed', 'procedure', 'function', 'to', 'do', 'downto', 'repeat',
    'until', 'then', 'asm', 'else');
  Operators: array[0..22] of string = ('@', 'not', '*', 'and', '/', 'shl',
    'shr', 'as', 'div', 'mod', 'or', 'xor', '-', '+', '=', '>', '<', '<>', '<=',
     '>=', 'is', 'in', ':=');

implementation

{ TFunction }

constructor TFunction.Create;
begin
  inherited;
  Parameters := TParameterList.Create;
  Parameters.Parent := Self;
  //ResultType := TType.Create;
end;

destructor TFunction.Destroy;
begin
  Parameters.Free;
//  ResultType.Free;
  inherited;
end;

{ TProgram }

constructor TProgram.Create;
begin
  Device := TDevice.Create;
  Modules := TObjectList.Create;
end;

destructor TProgram.Destroy;
begin
  Modules.Free;
  Device.Free;
end;

{ TConstant }


{ TConstantList }

destructor TConstantList.Destroy;
begin
  inherited;
end;

function TConstantList.Search(Name: string): TConstant;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (LowerCase(TConstant(Items[I]).Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := TConstant(Items[I]) else begin
    if Assigned(Parent.Parent) then Result := Parent.Parent.Constants.Search(Name)
      else begin
        Result := nil;
      end;
  end;
end;

{ TModule }

procedure TModule.Clear;
begin
  Variables.Clear;
  Constants.Clear;
  Functions.Clear;
  Code.Clear;
end;

constructor TModule.Create;
begin
  inherited;
  UsedModules := TUsedModuleList.Create;
end;

destructor TModule.Destroy;
begin
  UsedModules.Destroy;
  inherited;
end;

(*
procedure TCommonBlock.CheckReferences;
var
  I: Integer;
begin
  for I := 0 to Operations.Count - 1 do
  with TOperation(Operations[I]) do begin
    if (Instruction = inJump) or (Instruction = inConditionalJump) then
      TOperation(Operations[GotoAddress]).Referenced := True;
  end;
end;
*)

constructor TCommonBlock.Create;
begin
  Constants := TConstantList.Create;
  Constants.Parent := Self;
  Types := TTypeList.Create;
  Types.Parent := Self;
  Variables := TVariableList.Create;
  Variables.Parent := Self;
  Functions := TFunctionList.Create;
  Functions.Parent := Self;
  Code := TBeginEnd.Create;
  Code.Parent := Self;
  Code.CommonBlock := Self;
end;

destructor TCommonBlock.Destroy;
begin
  Constants.Destroy;
  Types.Destroy;
  Variables.Destroy;
  Functions.Destroy;
  Code.Destroy;
  inherited;
end;

{ TTypeList }

destructor TTypeList.Destroy;
begin
  inherited;
end;

function TTypeList.Search(Name: string): TType;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (LowerCase(TType(Items[I]).Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := TType(Items[I]) else Result := nil;

   if not Assigned(Result) and Assigned(Parent.Parent) then
     Result := Parent.Parent.Types.Search(Name);
end;

{ TVariableList }

destructor TVariableList.Destroy;
begin
  inherited;
end;

function TVariableList.Search(Name: string): TVariable;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (LowerCase(TVariable(Items[I]).Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := TVariable(Items[I]) else Result := nil;

  if not Assigned(Result) then begin
    if Assigned(Parent.Parent) then
      Result := Parent.Parent.Variables.Search(Name)
      else Result := nil;
  end;

  if not Assigned(Result) then begin
    if Assigned(Parent) and (Parent is TFunction) then
      Result := TFunction(Parent).Parameters.Search(Name)
      else Result := nil;
  end;

  if not Assigned(Result) then begin
    if Assigned(Parent.Parent) and (Parent.Parent is TFunction) then
      Result := TFunction(Parent.Parent).Parameters.Search(Name)
      else Result := nil;
  end;
end;

{ TVariable }

{ TType }


{ TFunctionList }

destructor TFunctionList.Destroy;
begin
  inherited;
end;

function TFunctionList.Search(Name: string): TFunction;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (LowerCase(TFunction(Items[I]).Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := TFunction(Items[I]) else begin
    if Assigned(Parent.Parent) then Result := Parent.Parent.Functions.Search(Name)
      else begin
        Result := nil;
      end;
  end;
end;

{ TExpression }

constructor TExpression.Create;
begin
  SubItems := TExpressionList.Create;
  SubItems.Count := 2;
  SubItems.OwnsObjects := False;
end;

destructor TExpression.Destroy;
begin
  SubItems.Free;
  inherited;
end;

procedure TExpression.Assign(Source: TExpression);
begin
  CommonBlock := Source.CommonBlock;
  NodeType := Source.NodeType;
  Method := Source.Method;
  Value := Source.Value;
  Associated := Source.Associated;
  OperatorName := Source.OperatorName;
  Variable := Source.Variable;
  SubItems.Assign(Source.SubItems);
end;

{ TOperationList }

destructor TOperationList.Destroy;
begin
  inherited;
end;

{ TOperation }

destructor TOperation.Destroy;
begin
  inherited;
  ExpressionTree.Free;
end;

{ TExpressionList }

destructor TExpressionList.Destroy;
begin
  inherited;
end;

{ TParameterList }

destructor TParameterList.Destroy;
begin
  inherited;
end;

function TParameterList.Search(Name: string): TParameter;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (LowerCase(TParameter(Items[I]).Name) <> LowerCase(Name)) do Inc(I);
  if I < Count then Result := TParameter(Items[I])
    else Result := nil;
end;

{ TBeginEnd }

procedure TBeginEnd.Clear;
begin

end;

constructor TBeginEnd.Create;
begin
  inherited;
  Commands := TCommandList.Create;
end;

destructor TBeginEnd.Destroy;
begin
  Commands.Free;
  inherited;
end;

{ TAssignment }

constructor TAssignment.Create;
begin
//  Source := TExpression.Create;
end;

destructor TAssignment.Destroy;
begin
  Source.Free;
  inherited;
end;

{ TWhileDo }

constructor TWhileDo.Create;
begin
  Condition := TExpression.Create;
  Command := TCommand.Create;
end;

destructor TWhileDo.Destroy;
begin
  Condition.Free;
  Command.Free;
  inherited;
end;

{ TCaseOfEnd }

constructor TCaseOfEnd.Create;
begin
  inherited;
  Branches := TObjectList.Create
end;

destructor TCaseOfEnd.Destroy;
begin
  Branches.Destroy;
  inherited;
end;

{ TIfThenElse }

constructor TIfThenElse.Create;
begin
  Condition := TExpression.Create;
end;

destructor TIfThenElse.Destroy;
begin
  Condition.Destroy;
  inherited Destroy;
end;

{ TFunctionCall }

constructor TFunctionCall.Create;
begin
  inherited;
  ParameterExpression := TObjectList.Create;
end;

destructor TFunctionCall.Destroy;
begin
  ParameterExpression.Free;
  inherited Destroy;
end;

{ TForToDo }

constructor TForToDo.Create;
begin
  inherited;
  Start := TExpression.Create;
  Stop := TExpression.Create;
end;

destructor TForToDo.Destroy;
begin
  Start.Free;
  Stop.Free;;
  inherited Destroy;
end;

{ TTypeRecord }

constructor TTypeRecord.Create;
begin
  inherited;
  Items := TObjectList.Create;
end;

destructor TTypeRecord.Destroy;
begin
  Items.Free;
  inherited Destroy;
end;

end.

